clc;                             % clear command window/console
clear;                           % clear workspace                                      
close all;                       % close all windows (plots)

import casadi.*

% problem settings
USE_REST_LENGTH    = 0;     % 0: don't use rest length; 1: use rest length
GROUND_CONSTRAINTS = 0;     % 0: only constraints on two ends of chain
                            % 1: linear ground constraints
                            % 2: convex quadratic constraint
                            % 3: nonconvex quadratic constraint
CUSTOM_INITIALIZATION = 0;  % 1: initialize to find different solution for nonconvex case       
                
% create empty optimization problem
opti = casadi.Opti();

% problem parameters
N = 40;
mi = 4/N;
Di = (70/40)*N;
g0 = 9.81;
L  = 1;

if USE_REST_LENGTH
    s = opti.variable(N-1,1);
    % length of spring in [m]
    Li = L/(N-1);
end

% definition of variables
y = opti.variable(N,1);
z = opti.variable(N,1);

% objective function
Vchain = 0;
for i = 1:N
    Vchain = Vchain + g0*mi*z(i);
end
if ~USE_REST_LENGTH
    for i=1:N-1
        Vchain = Vchain + 1/2*Di*((y(i)-y(i+1))^2 + (z(i)-z(i+1))^2);
    end
else
    for i=1:N-1
        Vchain = Vchain + 1/2*Di*s(i)^2;
    end
end

% pass objective to solver
opti.minimize(Vchain)

% constraints
opti.subject_to( y(1) == -2 )
opti.subject_to( z(1) ==  1 )
opti.subject_to( y(N) ==  2 )
opti.subject_to( z(N) ==  1 )

if USE_REST_LENGTH
    d = sqrt(diff(y).^2 + diff(z).^2) - Li;
    opti.subject_to( s >= 0);
    opti.subject_to( s >= d);
%     set to to avoid initializing at d at sqrt(0)
    y0 = linspace(-1,1,N);
    opti.set_initial(y,y0);
end

if GROUND_CONSTRAINTS == 1
    % linear constraints
    opti.subject_to(       z >= 0.5);
    opti.subject_to( z-0.1*y >= 0.5); 
elseif GROUND_CONSTRAINTS == 2
    % convex quadratic constraints
    opti.subject_to( z >= -0.2 + 0.1*y.^2); 
elseif GROUND_CONSTRAINTS == 3
    % non-convex quadratic constraints
    opti.subject_to( z >= -y.^2); 
end

% Setting the solver
opti.solver('ipopt')

if CUSTOM_INITIALIZATION
    opti.set_initial(y,2)
%     opti.set_initial(y,-2)
    opti.set_initial(z,-1)
end

sol = opti.solve();

% to see some more info
% disp(sol.stats)

if strcmp(sol.stats.return_status, 'Solve_Succeeded')
    
    Y = sol.value(y);
    Z = sol.value(z);
    
    figure;
    plot(Y,Z,'--or'); hold on;
    plot(-2,1,'xg','MarkerSize',10);
    plot(2,1,'xg','MarkerSize',10);
    xlabel('y'); ylabel('z');
    title('Optimal solution hanging chain')
    
    if GROUND_CONSTRAINTS == 1
        plot([-2 2],[0.5 0.5],':b');
        plot([-2 2],[0.3 0.7],':b');
    elseif GROUND_CONSTRAINTS == 2
        yi = -2:0.01:2;
        zi = -0.2 + 0.1*yi.^2;
        %zi = -1.3 + 0.5*yi.^2;
        plot(yi,zi,':b');
    elseif GROUND_CONSTRAINTS == 3
        yi = -2:0.01:2;
        zi = -yi.^2;
        plot(yi,zi,':b');
        axis([-2 2 -1 2])
    end
end